@@ -8,7 +8,7 @@ use base qw/Catalyst::Action/;
 sub execute {
     my $self = shift;
     my ( $controller, $c ) = @_;
-    $self->NEXT::execute( @_ );
+    $self->next::method( @_ );
     $c->res->header( 'X-Action-After', $c->stash->{after_message} );
 }
 
